package ru.dbotthepony.mc.otm.item

import it.unimi.dsi.fastutil.ints.Int2ObjectAVLTreeMap
import it.unimi.dsi.fastutil.ints.IntAVLTreeSet
import net.minecraft.ChatFormatting
import net.minecraft.core.Direction
import net.minecraft.nbt.CompoundTag
import net.minecraft.network.FriendlyByteBuf
import net.minecraft.network.chat.Component
import net.minecraft.world.item.*
import net.minecraft.world.level.Level
import net.minecraftforge.common.capabilities.Capability
import net.minecraftforge.common.capabilities.ForgeCapabilities
import net.minecraftforge.common.capabilities.ICapabilityProvider
import net.minecraftforge.common.util.LazyOptional
import net.minecraftforge.event.TickEvent
import net.minecraftforge.event.TickEvent.ServerTickEvent
import net.minecraftforge.network.NetworkEvent
import net.minecraftforge.registries.ForgeRegistries
import net.minecraftforge.registries.ForgeRegistry
import ru.dbotthepony.mc.otm.*
import ru.dbotthepony.mc.otm.capability.IMatteryEnergyStorage
import ru.dbotthepony.mc.otm.capability.MatteryCapability
import ru.dbotthepony.mc.otm.core.itemStackIterator
import ru.dbotthepony.mc.otm.core.nonEmpty
import ru.dbotthepony.mc.otm.core.*
import ru.dbotthepony.mc.otm.network.GenericNetworkChannel
import ru.dbotthepony.mc.otm.network.MatteryPacket
import ru.dbotthepony.mc.otm.network.packetHandled
import ru.dbotthepony.mc.otm.saveddata.SavedCountingMap
import ru.dbotthepony.mc.otm.saveddata.SavedMapDelegate
import java.util.function.Supplier

class QuantumBatteryItem : Item {
	private inner class Power(private val stack: ItemStack) : IMatteryEnergyStorage, ICapabilityProvider {
		private val resolver = LazyOptional.of { this }
		var delegate = SavedMapDelegate(ImpreciseFraction.ZERO)

		override fun <T : Any?> getCapability(cap: Capability<T>, side: Direction?): LazyOptional<T> {
			if (cap == ForgeCapabilities.ENERGY || cap == MatteryCapability.ENERGY) {
				return resolver.cast()
			}

			return LazyOptional.empty()
		}

		override fun extractEnergyOuter(howMuch: ImpreciseFraction, simulate: Boolean): ImpreciseFraction {
			return extractEnergyInner(howMuch, simulate)
		}

		override fun extractEnergyInner(howMuch: ImpreciseFraction, simulate: Boolean): ImpreciseFraction {
			if (howMuch.isNegative) {
				return ImpreciseFraction.ZERO
			}

			if (delegate.parent == null && isServerThread()) {
				determineQuantumLink()

				if (delegate.parent == null) {
					return ImpreciseFraction.ZERO
				}
			}

			if (isCreative) {
				val newEnergy = (delegate.value - howMuch).moreThanZero()
				val diff = delegate.value - newEnergy

				if (!simulate) {
					delegate.value = newEnergy
				}

				return diff
			}

			val newEnergy = (delegate.value - howMuch.coerceAtMost(throughput!!)).moreThanZero()
			val diff = delegate.value - newEnergy

			if (!simulate) {
				delegate.value = newEnergy
			}

			return diff
		}

		override fun receiveEnergyOuter(howMuch: ImpreciseFraction, simulate: Boolean): ImpreciseFraction {
			return receiveEnergyInner(howMuch, simulate)
		}

		override fun receiveEnergyInner(howMuch: ImpreciseFraction, simulate: Boolean): ImpreciseFraction {
			if (howMuch.isNegative) {
				return ImpreciseFraction.ZERO
			}

			if (delegate.parent == null && isServerThread()) {
				determineQuantumLink()

				if (delegate.parent == null) {
					return ImpreciseFraction.ZERO
				}
			}

			if (isCreative) {
				if (!simulate) {
					delegate.value += howMuch
				}

				return howMuch
			}

			val newEnergy = (delegate.value + howMuch.coerceAtMost(throughput!!)).coerceAtMost(capacity!!)
			val diff = newEnergy - delegate.value

			if (!simulate) {
				delegate.value = newEnergy
			}

			return diff
		}

		override val batteryLevel: ImpreciseFraction
			get() {
				if (delegate.parent == null) {
					determineQuantumLink()
				}

				if (isClientThread()) {
					return clientPowerMap[delegate.index] ?: delegate.value
				}

				return delegate.value
			}

		override val maxBatteryLevel: ImpreciseFraction
			get() = capacity ?: (batteryLevel + ImpreciseFraction.LONG_MAX_VALUE)

		override val missingPower: ImpreciseFraction
			get() = if (isCreative) ImpreciseFraction.LONG_MAX_VALUE else super.missingPower

		override fun canExtract(): Boolean {
			return true
		}

		override fun canReceive(): Boolean {
			return true
		}

		private fun determineQuantumLink() {
			if (delegate.parent == null && isServerThread()) {
				val existing = stack.tag?.getInt("link_id")

				if (existing == null) {
					val old = delegate
					delegate = saveData!!.factorize()
					delegate.value = old.value
					stack.tagNotNull["link_id"] = delegate.index
				} else {
					delegate = saveData?.computeIfAbsent(existing) ?: SavedMapDelegate(null, existing, delegate.value)
				}
			} else if (!isServerThread()) {
				// client ?
				val existing = stack.tag?.getInt("link_id") ?: return

				if (existing != delegate.index) {
					delegate = SavedMapDelegate(delegate.parent, existing, delegate.value)
				}
			}
		}

		fun determineQuantumLinkWeak() {
			if (delegate.parent == null && isServerThread()) {
				val existing = stack.tag?.getInt("link_id")

				if (existing != null) {
					delegate = saveData?.computeIfAbsent(existing) ?: SavedMapDelegate(null, existing, delegate.value)
				}
			} else if (!isServerThread()) {
				val existing = stack.tag?.getInt("link_id") ?: return

				if (existing != delegate.index) {
					delegate = SavedMapDelegate(delegate.parent, existing, delegate.value)
				}
			}
		}
	}

	val isCreative: Boolean
	val capacity: ImpreciseFraction?
	val throughput: ImpreciseFraction?

	val saveDataID: String

	val saveData: SavedCountingMap<SavedMapDelegate<ImpreciseFraction>>? get() {
		if (isServerThread()) {
			return MINECRAFT_SERVER.overworld().dataStorage.computeIfAbsent({
				SavedMapDelegate.makeMap(ImpreciseFraction::serializeNBT, ImpreciseFraction.Companion::deserializeNBT, ImpreciseFraction.ZERO).load(it)
			}, {
				SavedMapDelegate.makeMap(ImpreciseFraction::serializeNBT, ImpreciseFraction.Companion::deserializeNBT, ImpreciseFraction.ZERO)
			}, saveDataID) ?: throw NullPointerException("Unable to get save data for $this in ${MINECRAFT_SERVER.overworld()}")
		}

		return null
	}

	val clientPowerMap: Int2ObjectAVLTreeMap<ImpreciseFraction> by lazy {
		check(isClient) { "Invalid side" }
		Int2ObjectAVLTreeMap()
	}

	constructor(saveDataID: String) : super(Properties().stacksTo(1).rarity(Rarity.EPIC).tab(OverdriveThatMatters.INSTANCE.CREATIVE_TAB)) {
		isCreative = true
		capacity = null
		throughput = null
		this.saveDataID = "otm_$saveDataID".intern()
	}

	constructor(saveDataID: String, capacity: ImpreciseFraction, io: ImpreciseFraction) : super(Properties().stacksTo(1).tab(OverdriveThatMatters.INSTANCE.CREATIVE_TAB)) {
		isCreative = false
		this.capacity = capacity
		this.throughput = io
		this.saveDataID = "otm_$saveDataID".intern()
	}

	override fun initCapabilities(stack: ItemStack, nbt: CompoundTag?): ICapabilityProvider {
		return Power(stack)
	}

	override fun appendHoverText(
		itemStack: ItemStack,
		p_41422_: Level?,
		components: MutableList<Component>,
		p_41424_: TooltipFlag
	) {
		super.appendHoverText(itemStack, p_41422_, components, p_41424_)

		val power = itemStack.getCapability(MatteryCapability.ENERGY).orThrow() as Power

		components.add(TranslatableComponent("otm.item.quantum_description").withStyle(ChatFormatting.DARK_GRAY))

		if (isCreative) {
			components.add(TranslatableComponent("otm.item.quantum_battery.creative_power", power.batteryLevel.formatPower()).withStyle(ChatFormatting.GRAY))
			components.add(TranslatableComponent("otm.item.quantum_battery.creative").withStyle(ChatFormatting.DARK_GRAY))
			components.add(TranslatableComponent("otm.item.quantum_battery.creative2").withStyle(ChatFormatting.DARK_GRAY))
		} else {
			components.add(TranslatableComponent("otm.item.power.normal.storage", power.batteryLevel.formatPower(), capacity!!.formatPower()).withStyle(ChatFormatting.GRAY))
			components.add(
				TranslatableComponent(
				"otm.item.power.normal.throughput",
				throughput!!.formatPower(),
				throughput.formatPower()
			).withStyle(ChatFormatting.GRAY))
		}

		components.add(TranslatableComponent("otm.item.quantum_link_id", power.delegate.index).withStyle(ChatFormatting.DARK_GRAY))
	}

	companion object {
		fun readPacket(buff: FriendlyByteBuf): ChargePacket {
			return ChargePacket(
				(ForgeRegistries.ITEMS as ForgeRegistry<Item>).getValue(buff.readInt()) as QuantumBatteryItem,
				buff.readInt(),
				buff.readImpreciseFraction()
			)
		}

		fun tick(event: ServerTickEvent) {
			if (event.phase == TickEvent.Phase.END) {
				for (ply in event.server.playerList.players) {
					val networkedChannels = IntAVLTreeSet()

					for (item in ply.containerMenu.itemStackIterator().nonEmpty()) {
						if (item.item is QuantumBatteryItem) {
							val power = item.getCapability(MatteryCapability.ENERGY).orThrow() as Power

							power.determineQuantumLinkWeak()

							if (power.delegate.index < 0) {
								continue
							}

							if (networkedChannels.add(power.delegate.index)) {
								GenericNetworkChannel.send(ply, ChargePacket(item.item as QuantumBatteryItem, power.delegate.index, power.batteryLevel))
							}
						}
					}
				}
			}
		}
	}

	class ChargePacket(val type: QuantumBatteryItem, val channel: Int, val value: ImpreciseFraction) : MatteryPacket {
		override fun write(buff: FriendlyByteBuf) {
			buff.writeInt((ForgeRegistries.ITEMS as ForgeRegistry<Item>).getID(type))
			buff.writeInt(channel)
			buff.writeImpreciseFraction(value)
		}

		override fun play(context: Supplier<NetworkEvent.Context>) {
			context.packetHandled = true
			type.clientPowerMap[channel] = value
		}
	}
}
