package ru.dbotthepony.mc.otm.item

import net.minecraft.ChatFormatting
import net.minecraft.nbt.CompoundTag
import net.minecraft.network.chat.Component
import net.minecraft.world.item.Item
import net.minecraft.world.item.ItemStack
import net.minecraft.world.item.Rarity
import net.minecraft.world.item.TooltipFlag
import net.minecraft.world.level.Level
import net.minecraftforge.common.capabilities.ICapabilityProvider
import ru.dbotthepony.mc.otm.OverdriveThatMatters
import ru.dbotthepony.mc.otm.capability.EnergyProducerItem
import ru.dbotthepony.mc.otm.capability.MatteryCapability
import ru.dbotthepony.mc.otm.core.*

open class SingleUseBatteryItem(
	val storage: ImpreciseFraction,
	val throughput: ImpreciseFraction? = null,
	properties: Properties = Properties().stacksTo(1).tab(OverdriveThatMatters.INSTANCE.CREATIVE_TAB)
) : Item(properties) {
	private inner class Power(itemStack: ItemStack) : EnergyProducerItem(itemStack, storage, throughput) {
		init {
			if (!itemStack.tagNotNull.contains(NBT_KEY)) {
				batteryLevel = storage
			}
		}
	}

	private val throughputText = throughput?.let { TranslatableComponent("otm.item.power.output_only", it.formatPower()).withStyle(ChatFormatting.GRAY) }

	override fun appendHoverText(
		itemStack: ItemStack,
		p_41422_: Level?,
		list: MutableList<Component>,
		p_41424_: TooltipFlag
	) {
		super.appendHoverText(itemStack, p_41422_, list, p_41424_)

		list.add(SINGLE_USE)

		if (throughputText != null) {
			list.add(throughputText)
		}

		itemStack.getCapability(MatteryCapability.ENERGY).ifPresentK {
			list.add(TranslatableComponent(
				"otm.item.power.normal.storage",
				it.batteryLevel.formatPower(),
				it.maxBatteryLevel.formatPower()
			).withStyle(ChatFormatting.GRAY))
		}
	}

	override fun initCapabilities(stack: ItemStack, nbt: CompoundTag?): ICapabilityProvider {
		return Power(stack)
	}

	companion object {
		private val SINGLE_USE = TranslatableComponent("otm.battery.single_use").withStyle(ChatFormatting.GRAY)
	}
}

class ZPMItem : SingleUseBatteryItem(MAX_STORAGE, THROUGHPUT, Properties().stacksTo(1).rarity(Rarity.EPIC).tab(OverdriveThatMatters.INSTANCE.CREATIVE_TAB)) {
	override fun appendHoverText(
		itemStack: ItemStack,
		p_41422_: Level?,
		list: MutableList<Component>,
		p_41424_: TooltipFlag
	) {
		super.appendHoverText(itemStack, p_41422_, list, p_41424_)
		list.add(DESCRIPTION)
	}

	companion object {
		val MAX_STORAGE = ImpreciseFraction(200_000_000_000_000L)
		val THROUGHPUT = ImpreciseFraction(200_000_000L)

		private val DESCRIPTION = TranslatableComponent("item.${OverdriveThatMatters.MOD_ID}.zpm_battery.description").withStyle(ChatFormatting.DARK_GRAY)
	}
}
